/*
   NAME
   freedup -- search for duplicate files in one or more directory hierarchies
   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2007 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 9 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.*/

/* GNU freedup was written by Andreas Neuper <ANeuper@web.de>  */
#include "freedup.h"

/* Name this program was called with. */
char *program_name;


/* SYNOPSIS */
void usage(void)
{
   printf(  "USAGE:\t%s [options] [<tree> ...]\n\n", program_name );
   printf(  "\t-a\t   provide compatibility to freedups by William Stearns.[=-up]\n"
   	    "\t-c\t   count file space savings per linked file.\n"
	    "\t-d\t   requires the modification time stamps to be equal.\n" 
	    "\t-f\t   requires the path-stripped file names to be equal.\n"
	    "\t-g\t   requires groups of the files to be equal before linking.\n" 
	    "\t-h\t   shows this help. [other option are ignored]\n"
	    "\t-i\t   Decide in interactive mode individually, what to do with them.\n"
	    "\t-l\t   only allow hardlinks. No symlinks are established.\n" 
	    "\t-m <bytes> only consider larger files. (if applicable use: -o \"-size +#\")\n" 
	    "\t-n\t   do not really perform links [no action].\n" );
   printf(  "\t-o <opts>  pass an option string to the initially called find command.\n"
	    "\t-p\t   requires file permissions to be equal.\n" 
	    "\t-q\t   produces no output during the run (toggles -c and -v to off).\n" 
	    "\t-s\t   generate symlinks although some given paths are relative.\n"
	    "\t-u\t   requires users of the files to be equal before linking.\n" 
	    "\t-v\t   display shell commands to perform linking [verbose mode].\n"
	    "\t-w\t   only weak symbolic links allthough hardlinks might be possible.\n" 
	    "\t-#\t   disable fast comparison using hash strings [default uses MD5].\n" 
	    "\t-0\t   disable linking of empty files i.e. files of size 0.\n" 
	    "\t<tree>\t   any directory to scan for duplicate files recursively.\n\n" );
   printf(  "\tOptions are toggle switches. Their final state applies.\n"
	    "\tLater <tree> entries are linked to the earlier ones.\n"
	    "\tIf no <tree> is given, filenames are expected from standard input.\n"
	    "\tWhen standard input is used the option -o has no effect.\n"
	    "\tVersion " VERSION " by Andreas Neuper (c)2007.\n\n"
         );
   return;
}
   
/* OPTION DEFAULTS  */
	int showsave=0;	/* -c */
	int sametime=0;	/* -d */
	int samename=0;	/* -f */
	int samegrp =0;	/* -g */
	int hardlink=0;	/* -l */
	int minbytes=0;	/* -m */
	int testmode=0;	/* -n */
	int sameperm=0;	/* -p */
	int quietrun=0;	/* -q */
	int sameuser=0;	/* -u */
	int showmode=0;	/* -v */
	int weaklink=0;	/* -w */
	int interact=0; /* -i */
	int nonzero=0;	/* -0 */
	int usehash=1;	/* -# */
	int nosymlinks=0;	/* -s */
	char*findopta=NULL;	/* -o */

/* DESCRIPTION

   main() handles the command line parameters
	  scans for files to consider
	  builds a list of all files to consider
	  qsort()s the list of files
   comparison_by_size() is called for two files by qsort()
	  for equally sized files compares hash signature
	  comparison_by_content() when the same size & hash
   comparison_by_content() is only called by comparison_by_size()
	  reads and compares both files block by block
	  determines priority, i.e. later args count less
	  calls link_files() for identical files
   link_files() replaces the second file by a link to the first
	  determines whether both files are on the same fs
	  uses inode hard links on the same device number
	  uses symlinks if device numbers differ

   Additionally various helper functions are deployed.
*/

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
#include <assert.h>
#include <ctype.h>

void perror(const char *s);
#include <errno.h>
int errno;

/* internal counters */
/* link actions performed */
int totalinked=0;
int symlinked=0;
int deleted=0;
unsigned long long linksaved=0;

/* The number of regular files to scan. */
int finfosize=INITIALFINFOCOUNT;
int filecount=0;
int dupscount=0;

/* Time at start of execution.  */
time_t start_time;

/* Seconds between 00:00 1/1/70 and either one day before now
   (the default), or the start of today (if -daystart is given). */
time_t cur_day_start;

/* Status value to return to system. */
int exit_status;

/* The array pointer we use for sorting. */
finfo*file_info;

/* The array pointer we use for referencing duplicates in interactive mode. */
dinfo*dups_info;

#ifdef MD5SUM
int hashreject=0;
int hashread=0;

static void gethash(finfo*a)
{
    char buffer[MAXPATHLENGTH+MATCHCNT+4];
    FILE*file2hash;
    int n=0;
    hashread++;
    /*
     * this routine uses external shell commands
     */
    sprintf(buffer,MD5SUM " \"%s\" 2>/dev/null", a->name);
    file2hash = popen (buffer, "r");
    if(file2hash == 0)
    {
	perror(buffer);
	exit(-1);
    }
    fgets(buffer,MAXPATHLENGTH,file2hash);
    a->hash = calloc(MATCHCNT+2,sizeof(char));
    if( feof(file2hash) )
    {
	/*  hash was unsuccessful;
	 *  usually fails due to special characters in the file name
	 *  (this was at least the reason for redirecting errors)
	 *  write std dummy value;
	 */
        for(n=0;n<MATCHCNT;n++)
	   (a->hash)[n]='0';
	/*  it would be better (correct)
	 *  to link it to some temporary file,
	 *  to evaluate it, and to continue then
	 */
    } else
    {
	/*  if everything works fine
	 *  there should appear 32 characters
	 *  followed by two spaces and the file name
	 */
        assert(buffer[MATCHCNT]==' ');
        assert(buffer[MATCHCNT+1]==' ');
        if(0) printf("MD5=%s\n", buffer);
	/* store the hash value to avoid recalculation */
        strncpy(a->hash,buffer,MATCHCNT);
    }
    pclose(file2hash);
    /* the string was already initialised with zero
     * since I am somewhat paranoid ...
     */
    a->hash[MATCHCNT]=0;
    return;
}
#endif /* MD5SUM */

int statcpy(finfo*a)
{
    static struct stat*astat;
    astat=&(a->info);
    if(lstat(a->name, astat) != 0)
    {
	perror("lstat() failed while reading file statistics");
	return(-1);
    }
    return 0;
}


void linkfilesecure(char*aname,char*bname,int(*lnk)(const char *, const char *))
{
    char tmpfilename[MAXPATHLENGTH+16];
    static struct stat tstat;
    /* 
     * generate unique tmp file name
     * move file to tmp ;
     * later link it ;
     * finally remove tmp
     */
    if( !testmode )
    {
	strcpy(tmpfilename,bname);
	strcat(tmpfilename,".XXXXXX");
	if( mktemp(tmpfilename) == NULL )
	{
	    perror("There is no unique temporory file name.");
	}
	if(lstat(tmpfilename,&tstat) != 0)
	{
	    /*
	     * The errror needs not to be catched, since it is wanted
	     * that no file exists with the target name
	     */
	    rename(bname,tmpfilename);
	} else
	{
	    fprintf(stderr,"Not able to rename \"%s\" to \"%s\".",bname,tmpfilename);
	}
	lnk(aname,bname);
	if( lnk == symlink ) symlinked++;
	unlink( tmpfilename );
    }
    if( showmode)
    {
	printf("ln %s\"%s\" \"%s\"\n",( lnk == symlink )?"-s ":"",aname,bname);
    }
    return;
}

int interaction(finfo*a, finfo*b)
{
    int n=0,m=0;

    /* merge or create dups_info entries */
    if( (a->ident == NULL) && (b->ident == NULL) )
    {
	dups_info[dupscount].name = (char**)calloc(32,sizeof(char**));
	dups_info[dupscount].name[0] = a->name;
	dups_info[dupscount].name[1] = b->name;
	/* other pointers are NULL, since calloc sets to zero */
	a->ident = b->ident = &dups_info[dupscount];
	dupscount++;
    } else
    {
	if( a->ident == NULL )
	{
	    /*clean up tangling references */
	    if( (b->ident->replace!=NULL) )
	    {
		b->ident = b->ident->replace;
	    }
	    /* simply inherit reference list and append the new filename */
	    a->ident = b->ident;
	    for(n=0;b->ident->name[n]!=NULL && n<32;n++);
	    b->ident->name[n] = a->name;
	} else
	if( b->ident == NULL )
	{
	    /*clean up tangling references */
	    if( (a->ident->replace!=NULL) )
	    {
		a->ident = a->ident->replace;
	    }
	    /* simply inherit reference list and append the new filename */
	    b->ident = a->ident;
	    for(n=0; a->ident->name[n]!=NULL && n<32;n++);
	    a->ident->name[n] = b->name;
	} else
	{
	    /*clean up tangling references */
	    while( ((a->ident->replace)!=NULL) )
	    {
		a->ident = a->ident->replace;
	    }
	    while( ((b->ident->replace)!=NULL) )
	    {
		b->ident = b->ident->replace;
	    }
	    /* merge the reference lists */
	    assert( a->ident->name!=NULL );
	    assert( b->ident->name!=NULL );
	    if( a->ident != b->ident )
	    {
		for(n=0;a->ident->name[n]!=NULL && n< 32   ;n++);
		for(m=0;b->ident->name[m]!=NULL && m<(32-n);m++) 
			a->ident->name[n+m] = b->ident->name[m];
		assert((n+m)<32);
		free(b->ident->name);
		b->ident->name = NULL;
		/* avoid active search for tangling references */
		b->ident->replace = a->ident;
		b->ident = a->ident;
	    }
	}
    }
    return 0;
}

int dupinfo_entry(dinfo*a)
{
    static struct stat astat;
    static int n,m,min,max,cnt,source;
    for( n=0; a->name[n]!=NULL; n++ )
    {
	if( lstat(a->name[n], &astat) != 0 )
	{
	    /* ignore this file, probably deleted */
	    continue;
	}
	if( testmode )
	{
	    printf( "%s\n", a->name[n] );
	}
	else
	{
	    printf( " %d : %c [%04x:%06x:%04o:%01d] %s\n",
		n, n+(n?'a':'A'),
		(int)astat.st_dev,
		(int)astat.st_ino,
		(int)S_PERM(astat.st_mode),
		(int)astat.st_nlink,
		a->name[n]);
	}
    }
    if( !testmode )
    {
        static char buffer[MAXPATHLENGTH],target[32];
	/* 
	 * analyse and execute the answer
	 */
	printf("\tDelete on number, Link on letter, Symlink on Capital (first is source)\t");
	fflush(stdout);
	fgets(buffer, MAXPATHLENGTH, stdin);
	strncpy(target,"               ",n);
	target[n]=0;
	source=-1;
	for(m=0;buffer[m]!=0 && m<MAXPATHLENGTH;m++)
	{
	    if( isalpha(buffer[m]) && source==-1 )
	    {   source = toupper(buffer[m]) - 'A';
		target[source] = 'Q';
	    } else {
	    if( isupper(buffer[m]) ) { max++; target[(buffer[m]-'A')]='S'; }
	    if( islower(buffer[m]) ) { min++; target[(buffer[m]-'a')]='L'; } }
	    if( isdigit(buffer[m]) ) { cnt++; target[(buffer[m]-'0')]='D'; }
	}
	if( (source==-1) && (max+min>0) ) return 1; /* all other cases are treated as valid */
        if( (source!=-1) && ((a->name[source])[0] != '/') )
	{
	    printf("DANGER: Resulting symlinks can be invalid due to relative pathname!");
	}
	/* 
	 * tell user what we will do
	 */
	puts("");
	for( n=0; a->name[n]!=NULL; n++ )
	{
	    if( lstat(a->name[n], &astat) != 0 )
	    {
		/* ignore this file, probably deleted */
		continue;
	    }
	    switch(target[n])
	    {
		case 'S':  if( showmode ) printf("\t$ "); else
			   printf( "\t$ ln -s \"%s\" \"%s\"\n", a->name[source], a->name[n] );
			   linkfilesecure(a->name[source],a->name[n],symlink);
			   break;
		case 'L':  if( showmode ) printf("\t$ "); else
			   printf( "\t$ ln \"%s\" \"%s\"\n", a->name[source], a->name[n] );
			   linkfilesecure(a->name[source],a->name[n],link);
			   break;
		case 'D':  printf( "\t$ rm -f %s\n", a->name[n] );
			   unlink( a->name[n] );
			   break;
		case 'Q':  break;
		case ' ':  printf( "\t$ #keep %s\n", a->name[n] );
			   break;
		default :  break;
	    }
	}
    }
    return 0;
}

void print_dupinfo(void)
{
    int n=0;
    for(n=0;n<dupscount;n++)
    {
	if(dups_info[n].name!=NULL)
	{
	    if(!testmode)
		printf("Del:Lnk [dev:i-node:perm:L] <filename>\n");
		if(dupinfo_entry(&dups_info[n])!=0) n--;
		printf("\n");
	}
    }
}

void 
link_files(finfo*a, finfo*b)
{
#ifdef DEBUG
    fprintf(stderr,"testmode=%d,showmode=%d,weaklink=%d,hardlink=%d,nosymlinks=%d,showsave=%d\n",
	testmode,showmode,weaklink,hardlink,nosymlinks,showsave);
#endif /*DEBUG*/
    if( S_ISLNK(a->info.st_mode) || S_ISLNK(b->info.st_mode) )
    {
        /* either one is already symlinked
	 * this avoids building up symlink chains (in case of a)
	 * or alternating symlinks during sort (in case of b)
	 */
	return;
    }
    if(nosymlinks && (a->info.st_dev != b->info.st_dev) )
    {
        /* when on different devices and symlinks are not allowed
	 * do not link, i.e. no also not prepare to link.
	 * This needs to be catched before any execution.
	 */
	return;
    }
    if(nonzero && (a->info.st_size == 0) )
    {
        /* when only non-empty files are allowed to be linked
	 * do not link, when they are identical and has size zero.
	 * This needs to be catched before any execution.
	 */
	return;
    }
    if( interact )
    {
	interaction(a,b);
	return;
    }
    if( !weaklink && ( a->info.st_dev == b->info.st_dev ) )
    {
	/* 
	 * either st_dev or st_ino are different,
	 * was checked earlier prior to call
	 */
	linkfilesecure(a->name,b->name,link);
	statcpy(b);
    } else
    {
	assert( weaklink || !nosymlinks );
        if(!hardlink )
	{
	    linkfilesecure(a->name,b->name,symlink);
	    statcpy(b);
	}
    }
    if( showsave ) 
    {
	printf("Linking \"%s\" saves %d bytes\n",
		basename(b->name), 
		(int)(b->info).st_size);
    }
    linksaved+=(int)(b->info).st_size;
    totalinked++;
    return;
}

int comparison_by_content(const void*a,const  void*b)
{
    FILE*afile=NULL,
	*bfile=NULL;
    char ablock[BLOCK+2],
	 bblock[BLOCK+2];
    int  aread,
	 bread,
	 result;
    assert(a!=NULL);
    assert(b!=NULL);
    afile=fopen(((finfo*)a)->name,"r");
    if( afile == NULL )
    {
	/* perror("fopen() failed - file was probably deleted"); */
	return(-1);
    }
    bfile=fopen(((finfo*)b)->name,"r");
    if( bfile == NULL )
    {
	/* perror("fopen() failed - file was probably deleted"); */
	return(1);
    }
    while(!(feof(afile)||feof(bfile)))
    {
	aread=fread(ablock,BLOCK,1,afile);
	assert((aread!=0)||feof(afile));
	bread=fread(bblock,BLOCK,1,bfile);
	assert((bread!=0)||feof(bfile));
	assert(aread==bread);
	result=memcmp(ablock,bblock,aread);
	if( result!=0 )
	{
	    fclose(afile);
	    fclose(bfile);
	    return(result);
        }
    }
    /* decide which has higher prio by
     * 1) collect links where more hardlinks exist
     * 2) link to the one earlier named in command line
     * 3) arbitrary /# inactive: shorter path name #/
     * NOT: shorter path length /+/ may result in alternating linkage
     * the one with lower priority will be deleted
     */
    if(((finfo*)a)->info.st_nlink > ((finfo*)b)->info.st_nlink )
    {
	/* We should elimate unlinked files and
	 * keep more often linked files as collectors
	 */
	link_files(((finfo*)a),((finfo*)b));
    } else
    {
	if(((finfo*)a)->info.st_nlink < ((finfo*)b)->info.st_nlink )
	{
	    link_files(((finfo*)b),((finfo*)a));
	} else
	{
	    if(((finfo*)a)->prio<((finfo*)b)->prio)
	    {
		link_files(((finfo*)a),((finfo*)b));
	    } else
	    {
		if(((finfo*)a)->prio>((finfo*)b)->prio)
		{
		    link_files(((finfo*)b),((finfo*)a));
		} else
		{
/**/
/* it did not work well to give priority within one file system tree	*/
/*	    if(strlen(((finfo*)a)->name)<strlen(((finfo*)b)->name))	*/
/**/
		    link_files(((finfo*)a),((finfo*)b));
		}
	    }
	}
    }
    fclose(afile);
    fclose(bfile);
    return(0);
}

static int comparison_by_size(const void*a, const void*b)
{
    static int sizediff;
    assert(a!=NULL);
    assert(b!=NULL);
    sizediff = (( (((finfo*)a)->info.st_size) 
                - (((finfo*)b)->info.st_size) ));
    /*
     * avoid work when not desired
     * follow the minbytes directive
     */
    if((minbytes > (((finfo*)a)->info.st_size))
    || (minbytes > (((finfo*)b)->info.st_size)))
      {
        return(sizediff);
      }
    /*
     * up to here it is obvious:
     * - different size different file
     * - same size -> differs file?
     */
    if(sizediff == 0)
    {
	/* exclude weaklinks from linking, must not result in endless link loops */
	if( (((finfo*)a)->info.st_mode & S_IFMT) != S_IFREG 
	 || (((finfo*)b)->info.st_mode & S_IFMT) != S_IFREG )
	    /* same size but no linking allowed */
	    return 0;
	/* compare the path-less file names
	 * if the samename directive (-f) requires this */
	if( strcmp(basename(((finfo*)a)->name),basename(((finfo*)b)->name)) 
	 && (samename==1) )
	    /* different names do not allow linking */
	    return strlen(basename(((finfo*)a)->name))-strlen(basename(((finfo*)b)->name));
	/* check whether the user entries match (-u) */
	if( ((((finfo*)a)->info.st_uid)!=(((finfo*)b)->info.st_uid)) 
	 &&  (sameuser==1) )
	    /* different user ids do not allow linking */
	    return ((((finfo*)a)->info.st_uid)-(((finfo*)b)->info.st_uid));
	/* check whether the group entries match (-g) */
	if( ((((finfo*)a)->info.st_gid)!=(((finfo*)b)->info.st_gid)) 
	 &&  (samegrp ==1) )
	    /* different group ids do not allow linking */
	    return ((((finfo*)a)->info.st_gid)-(((finfo*)b)->info.st_gid));
	/* check whether the permissions match (-p) */
	if( (((((finfo*)a)->info.st_mode)&(ALLPERMS))!=((((finfo*)b)->info.st_mode)&(ALLPERMS))) 
	 && (sameperm==1) )
	    /* different permissions do not allow linking */
	    return 0;
	/* check whether the modification dates match (-d) */
	if( ((((finfo*)a)->info.st_mtime)!=(((finfo*)b)->info.st_mtime)) 
	 && (sametime==1) )
	    /* different modification time do not allow linking */
	    return((((finfo*)a)->info.st_mtime)-(((finfo*)b)->info.st_mtime)); 
	/* check whether already hardlinked */
	if( ((((finfo*)a)->info.st_ino != ((finfo*)b)->info.st_ino))
	 || ((((finfo*)a)->info.st_dev != ((finfo*)b)->info.st_dev)) )
	{
#ifdef MD5SUM
	    if( usehash )
	    {
		if((((finfo*)a)->hash)==NULL) gethash((finfo*)a);
		if((((finfo*)b)->hash)==NULL) gethash((finfo*)b);
		assert(((finfo*)a)->hash != NULL);
		assert(((finfo*)b)->hash != NULL);
		if( strncmp( (((finfo*)a)->hash), (((finfo*)b)->hash), MATCHCNT )!=0 )
		{
		    hashreject++;
		    return( strncmp( (((finfo*)a)->hash), (((finfo*)b)->hash), MATCHCNT ) );
		} else
		{
		    comparison_by_content(a,b);
		}
	    } else
#endif
	    {
		comparison_by_content(a,b);
	    }
	}
    }
    return(sizediff);
}

void print_infolist(void)
{
    int n;
    for(n=0;n<filecount;n++)
    {
        printf("%9d %x:%05d %s\n",
	  (int)file_info[n].info.st_size,
	  (int)file_info[n].info.st_dev,
	  (int)file_info[n].info.st_ino,
	  file_info[n].name);
    }
}

void print_duplist(void)
{
    int n,m,identicalcount=0,alreadylinked=0;
    for(n=0;n<filecount;n++)
    {
        for(m=1;n+m<filecount && (file_info[n].info.st_size==file_info[n+m].info.st_size);m++)
        {
	    identicalcount=m+1;
	    if( (file_info[n].info.st_dev  == file_info[n+m].info.st_dev)
	    &&  (file_info[n].info.st_ino  == file_info[n+m].info.st_ino) )
	    {
		alreadylinked+=1;
	    }
if(1)	    printf("%01d\t%01d\t",identicalcount,alreadylinked);
	}
	if( (identicalcount==m) && (identicalcount==alreadylinked+1) )
	{
	    identicalcount=0;
	} 
 	alreadylinked=0;
	{
	printf("%9d %x:%05d %s\n",
	  (int)file_info[n].info.st_size,
	  (int)file_info[n].info.st_dev,
	  (int)file_info[n].info.st_ino,
	  file_info[n].name);
	    identicalcount--;
	}
    }
}

void printstats(void)
{
    if(deleted) fprintf(stderr,"%d Files have been%s deleted\n",deleted,testmode?" marked to be":"");
    fprintf(stderr,"%d file%s of %d %sreplaced by links.\nThe total size of replac%s files %s %lld bytes.\n",
	totalinked,(totalinked==1)?"":"s",filecount,testmode?"will be ":"",
	testmode?"eable":"ed", testmode?"is":"was", linksaved);
#ifdef MD5SUM
    if( usehash )
    fprintf(stderr,"md5 hash algorithm had to read %d files to avoid %d file comparisons.\n",hashread,hashreject);
#endif /* MD5SUM */
}


int readtree( finfo**fin, int*total, const int filecount, FILE*in)
/*
 * "total" will be modified
 * "fin" might be modified
 * "in" is read until EOF
 */
{
    char buffer[MAXPATHLENGTH];
    int  count;

    for(count=filecount;!feof(in);count++)
    {
	fgets(buffer,MAXPATHLENGTH,in);
	if( !feof(in) && (strlen(buffer)>1) )
	{
	    (*fin)[count].prio=count;
	    (*fin)[count].name=(char*)malloc(strlen(buffer)+2);
	    if( (*fin)[count].name == NULL )
	    {
		perror("malloc() for buffer in file info structure");
		exit(-1);
	    }
	    assert(buffer[strlen(buffer)-1]=='\n');
	    /* fails if MAXPATHLENGTH is exceeded */
	    buffer[strlen(buffer)-1]=0;
	    strcpy((*fin)[count].name,buffer);
	    (*fin)[count].ident=NULL;
	    if( statcpy(&((*fin)[count]))!=0
	    /*
	     * Drop symbolic links here
	     */
	    ||  !S_ISREG((*fin)[count].info.st_mode) )
	    {
		free((*fin)[count].name);
		count--;
		continue;
	    }
	    if( count > (*total)-2 )
	    {   /*
		 * expand file_info array to take more members
		 */
		(*total) = (*total) + INITIALFINFOCOUNT;
		(*fin) = realloc((*fin), sizeof(finfo) * ( (*total) +4 ) );
		if( (*fin) == NULL )
		{
		    perror("realloc() to extend file info structure");
		    exit(-1);
		}
	    }
	} else count--;
    }
    return count;
}

int
main (int argc, char *argv[])
{
  int c,i,lastlinked=0,remember=0;
  FILE*liste;
  char buffer[MAXPATHLENGTH];
  program_name = argv[0];

#ifdef HAVE_SETLOCALE
  setlocale (LC_ALL, "");
#endif
#ifdef DEBUG
  /* Locale is not used in current implementation
   * Time is not used in current implementation
   */
  start_time = time (NULL);
  if(!quietrun) printf ("cur_day_start = %s", ctime (&cur_day_start));
#endif /* DEBUG */
  exit_status = 0;

    while ((c = getopt(argc, argv, "acdfghilm:no:pqsuvw#0")) != EOF) 
    {
        switch (c)
	{
	  case 'a': sameperm   = sameuser = samegrp = 1; break;
	  case 'c': showsave   = 1-showsave;   break;
	  case 'd': sametime   = 1-sametime;   break;
	  case 'f': samename   = 1-samename;   break;
	  case 'g': samegrp    = 1-samegrp ;   break;
	  case 'h': usage();                 exit(0);
	  case 'i': interact   = 1-interact;   break;
	  case 'l': hardlink   = 1-hardlink;   break;
	  case 'm': minbytes   = atoi(optarg); break;
	  case 'n': testmode   = 1-testmode;   break;
	  case 'o': findopta   = calloc( strlen(optarg)+2, sizeof(char) );
		    strcpy( findopta, optarg );break;
	  case 'p': sameperm   = 1-sameperm;   break;
	  case 'q': quietrun   = 1-quietrun;   showsave = showmode = 0; break;
	  case 's': nosymlinks = 1-nosymlinks; break;
	  case 'u': sameuser   = 1-sameuser;   break;
	  case 'v': showmode   = 1-showmode;   break;
	  case 'w': weaklink   = 1-weaklink;   break;
	  case '#': usehash    = 1-usehash;    break;
	  case '0': nonzero    = 1-nonzero;    break;
	  default:
		fprintf(stderr,"wrong option: \"%c\"\n",c);
		usage();
		exit(-1);
	}
    }
  /* 
   * Collect files to scan.
   * Read them and fill in stat now
   */
  file_info = (finfo*)calloc(finfosize+4,sizeof(finfo));
  if( file_info == NULL )
  {
      perror("calloc() to provide file info structure");
      return(-1);
  }
  if( argc == optind )
    {	/*
	 * Read file names (not dir names) from stdin
	 */
	if(!quietrun) fprintf(stderr,"Taking file names from stdin\n\n");
	filecount = readtree( &file_info, &finfosize, 0, stdin );
	/*
	 * Now we will skip the following for-loop
	 */
    } else
    {
      for (i=optind,filecount=0;i < argc;i++)
	{
	  if( argv[i][0] != '/' && (argc-optind) > 1 )
	    {
	      if(!quietrun) fprintf(stderr,"WARNING: disabling symbolic links due to relative paths\n");
	      nosymlinks=1;
	    }
	  sprintf(buffer,"find \"%s\" -type f %s -print", argv[i],(findopta==NULL)?"":findopta);
	  liste = popen (buffer, "r");
	  if(liste == 0)
	  {
	      perror(buffer);
	      return(-1);
	  }
	  filecount = readtree( &file_info, &finfosize, filecount, liste );
	  pclose(liste);
	}
    }
    if(interact)
    {
	/* worst case estimate is that all files are duplicate */
	dups_info = (dinfo*)calloc( (filecount>>1)+2, sizeof(dinfo) );
	if( dups_info == NULL )
	{
	    perror("alloc() does not provide info structure for duplicate files");
	    exit(-1);
	}
    }
    if(!quietrun) fprintf(stderr,"%d files to investigate\n\n",filecount);
    do
    {
	remember=totalinked-lastlinked;
        lastlinked=totalinked;
        qsort( file_info, filecount, sizeof(finfo), comparison_by_size );
#ifdef DEBUG
        if(!quietrun) 
	{
	    fprintf(stderr,"sorting loop ended with %d:%d\n",lastlinked,totalinked);
            print_infolist();
	}
#endif /* DEBUG */
    } while(lastlinked<totalinked && remember!=(totalinked-lastlinked) && ! testmode);
    if(!quietrun&&!interact) printstats();
    if(interact)  print_dupinfo();
    exit(exit_status);
}
